details.toggleable-item {
  user-select: none !important;
  padding: 0 !important;
}

details.toggleable-item .menu {
  margin: 4px 0 10px !important;
}

details.toggleable-item summary {
  display: flex;
  justify-content: space-between;
  align-items: center;
  padding: 0.92857143em 1.14285714em;
}

details.toggleable-item summary::marker, /* Chrome, Edge, Firefox */
details.toggleable-item summary::-webkit-details-marker /* Safari */ {
  display: none;
}

details.toggleable-item summary::after {
  transition: transform 0.25s ease;
  content: "";
  width: 14px;
  height: 14px;
  mask-size: cover;
  -webkit-mask-size: cover;
  mask-image: var(--octicon-chevron-right);
  -webkit-mask-image: var(--octicon-chevron-right);
  background: currentcolor;
  border: 1px solid var(--color-body); /* workaround https://bugzilla.mozilla.org/show_bug.cgi?id=1671784 */
}

details.toggleable-item[open] summary::after {
  transform: rotate(90deg);
}
