@import "../chroma/light.css";
@import "../codemirror/light.css";
@import "../markup/light.css";

:root {
  --steel-900: #10161d;
  --steel-850: #131a21;
  --steel-800: #171e26;
  --steel-750: #1d262f;
  --steel-700: #242d38;
  --steel-650: #2b3642;
  --steel-600: #374351;
  --steel-550: #445161;
  --steel-500: #515f70;
  --steel-450: #5f6e80;
  --steel-400: #6d7d8f;
  --steel-350: #7c8c9f;
  --steel-300: #8c9caf;
  --steel-250: #9dadc0;
  --steel-200: #aebed0;
  --steel-150: #c0cfe0;
  --steel-100: #d2e0f0;
  --zinc-50: #fafafa;
  --zinc-100: #f4f4f5;
  --zinc-150: #ececee;
  --zinc-200: #e4e4e7;
  --zinc-250: #dcdce0;
  --zinc-300: #d4d4d8;
  --zinc-350: #babac1;
  --zinc-400: #a1a1aa;
  --zinc-450: #898992;
  --zinc-500: #71717a;
  --zinc-550: #61616a;
  --zinc-600: #52525b;
  --zinc-650: #484850;
  --zinc-700: #3f3f46;
  --zinc-750: #333338;
  --zinc-800: #27272a;
  --zinc-850: #1f1f23;
  --zinc-900: #18181b;
  --color-primary: #c2410c;
  --color-primary-contrast: #ffffff;
  --color-primary-dark-1: #c2410c;
  --color-primary-dark-2: #9a3412;
  --color-primary-dark-3: #9a3412;
  --color-primary-dark-4: #7c2d12;
  --color-primary-dark-5: #7c2d12;
  --color-primary-dark-6: #7c2d12;
  --color-primary-dark-7: #7c2d12;
  --color-primary-light-1: #ea580c;
  --color-primary-light-2: #f97316;
  --color-primary-light-3: #fb923c;
  --color-primary-light-4: #fdba74;
  --color-primary-light-5: #fed7aa;
  --color-primary-light-6: #ffedd5;
  --color-primary-light-7: #fff7ed;
  --color-primary-alpha-10: #c2410c19;
  --color-primary-alpha-20: #c2410c33;
  --color-primary-alpha-30: #c2410c4b;
  --color-primary-alpha-40: #c2410c66;
  --color-primary-alpha-50: #c2410c80;
  --color-primary-alpha-60: #c2410c99;
  --color-primary-alpha-70: #c2410cb3;
  --color-primary-alpha-80: #c2410ccc;
  --color-primary-alpha-90: #c2410ce1;
  --color-primary-hover: var(--color-primary-dark-2);
  --color-primary-active: var(--color-primary-dark-4);
  --color-secondary: var(--zinc-200);
  --color-secondary-dark-1: var(--zinc-200);
  --color-secondary-dark-2: var(--zinc-300);
  --color-secondary-dark-3: var(--zinc-300);
  --color-secondary-dark-4: var(--zinc-400);
  --color-secondary-dark-5: var(--zinc-400);
  --color-secondary-dark-6: var(--zinc-500);
  --color-secondary-dark-7: var(--zinc-500);
  --color-secondary-dark-8: var(--zinc-600);
  --color-secondary-dark-9: var(--zinc-600);
  --color-secondary-dark-10: var(--zinc-700);
  --color-secondary-dark-11: var(--zinc-700);
  --color-secondary-dark-12: var(--zinc-800);
  --color-secondary-dark-13: var(--zinc-800);
  --color-secondary-light-1: var(--zinc-200);
  --color-secondary-light-2: var(--zinc-100);
  --color-secondary-light-3: var(--zinc-100);
  --color-secondary-light-4: var(--zinc-50);
  --color-secondary-alpha-10: #d4d4d819;
  --color-secondary-alpha-20: #d4d4d833;
  --color-secondary-alpha-30: #d4d4d84b;
  --color-secondary-alpha-40: #d4d4d866;
  --color-secondary-alpha-50: #d4d4d880;
  --color-secondary-alpha-60: #d4d4d899;
  --color-secondary-alpha-70: #d4d4d8b3;
  --color-secondary-alpha-80: #d4d4d8cc;
  --color-secondary-alpha-90: #d4d4d8e1;
  --color-secondary-hover: var(--color-secondary-dark-2);
  --color-secondary-active: var(--color-secondary-dark-4);
  /* console colors - used for actions console and console files */
  --color-console-fg: #eeeff2;
  --color-console-fg-subtle: #959cab;
  --color-console-bg: #1f212b;
  --color-console-border: #383c47;
  --color-console-hover-bg: #ffffff16;
  --color-console-active-bg: #454a57;
  --color-console-menu-bg: #383c47;
  --color-console-menu-border: #5c6374;
  /* colors */
  --color-red: #dc2626;
  --color-orange: #ea580c;
  --color-yellow: #ca8a04;
  --color-olive: #91a313;
  --color-green: #15803d;
  --color-teal: #0d9488;
  --color-blue: #2563eb;
  --color-violet: #7c3aed;
  --color-purple: #9333ea;
  --color-pink: #db2777;
  --color-brown: #a47252;
  --color-grey: #4b5563;
  --color-black: #000000;
  /* light variants */
  --color-red-light: #ef4444;
  --color-orange-light: #f97316;
  --color-yellow-light: #eab308;
  --color-olive-light: #839311;
  --color-green-light: #16a34a;
  --color-teal-light: #14b8a6;
  --color-blue-light: #3b82f6;
  --color-violet-light: #8b5cf6;
  --color-purple-light: #a855f7;
  --color-pink-light: #ec4899;
  --color-brown-light: #94674a;
  --color-grey-light: #6b7280;
  --color-black-light: #181818;
  /* dark 1 variants - produced via Sass scale-color(color, $lightness: -10%) */
  --color-red-dark-1: #c82020;
  --color-orange-dark-1: #d34f0b;
  --color-yellow-dark-1: #b67c04;
  --color-olive-dark-1: #839311;
  --color-green-dark-1: #137337;
  --color-teal-dark-1: #0c857a;
  --color-blue-dark-1: #1554e0;
  --color-violet-dark-1: #6a1feb;
  --color-purple-dark-1: #8519e7;
  --color-pink-dark-1: #c7216b;
  --color-brown-dark-1: #94674a;
  --color-black-dark-1: #000000;
  /* dark 2 variants - produced via Sass scale-color(color, $lightness: -20%) */
  --color-red-dark-2: #b21d1d;
  --color-orange-dark-2: #bb460a;
  --color-yellow-dark-2: #a26e03;
  --color-olive-dark-2: #74820f;
  --color-green-dark-2: #116631;
  --color-teal-dark-2: #0a766d;
  --color-blue-dark-2: #124bc7;
  --color-violet-dark-2: #5c14d8;
  --color-purple-dark-2: #7715cf;
  --color-pink-dark-2: #b11d5f;
  --color-brown-dark-2: #835b42;
  --color-black-dark-2: #000000;
  /* ansi colors used for actions console and console files */
  --color-ansi-black: #1f2326;
  --color-ansi-red: #cc4848;
  --color-ansi-green: #87ab63;
  --color-ansi-yellow: #cc9903;
  --color-ansi-blue: #3a8ac6;
  --color-ansi-magenta: #d22e8b;
  --color-ansi-cyan: #00918a;
  --color-ansi-white: var(--color-console-fg-subtle);
  --color-ansi-bright-black: #46494d;
  --color-ansi-bright-red: #d15a5a;
  --color-ansi-bright-green: #93b373;
  --color-ansi-bright-yellow: #eaaf03;
  --color-ansi-bright-blue: #4e96cc;
  --color-ansi-bright-magenta: #d74397;
  --color-ansi-bright-cyan: #00b6ad;
  --color-ansi-bright-white: var(--color-console-fg);
  /* other colors */
  --color-gold: #b1983b;
  --color-white: #ffffff;
  --color-diff-removed-word-bg: #fca5a5;
  --color-diff-added-word-bg: #86efac;
  --color-diff-removed-row-bg: #fee2e2;
  --color-diff-moved-row-bg: #fef9c3;
  --color-diff-added-row-bg: #dcfce7;
  --color-diff-removed-row-border: #fca5a5;
  --color-diff-moved-row-border: #fde047;
  --color-diff-added-row-border: #86efac;
  --color-diff-inactive: var(--zinc-100);
  --color-error-border: #fecaca;
  --color-error-bg: #fee2e2;
  --color-error-bg-active: #fca5a5;
  --color-error-bg-hover: #fecaca;
  --color-error-text: #7f1d1d;
  --color-success-border: #bbf7d0;
  --color-success-bg: #dcfce7;
  --color-success-text: #14532d;
  --color-warning-border: #fde047;
  --color-warning-bg: #fef3c7;
  --color-warning-text: #78350f;
  --color-info-border: #bae6fd;
  --color-info-bg: #e0f2fe;
  --color-info-text: #0c4a6e;
  --color-red-badge: #b91c1c;
  --color-red-badge-bg: #b91c1c22;
  --color-red-badge-hover-bg: #b91c1c44;
  --color-green-badge: #16a34a;
  --color-green-badge-bg: #16a34a22;
  --color-green-badge-hover-bg: #16a34a44;
  --color-yellow-badge: #ca8a04;
  --color-yellow-badge-bg: #ca8a0422;
  --color-yellow-badge-hover-bg: #ca8a0444;
  --color-orange-badge: #ea580c;
  --color-orange-badge-bg: #ea580c22;
  --color-orange-badge-hover-bg: #ea580c44;
  /* target-based colors */
  --color-body: #fff;
  --color-box-header: var(--zinc-100);
  --color-box-body: var(--zinc-50);
  --color-box-body-highlight: var(--zinc-200);
  --color-text-dark: #000;
  --color-text: var(--zinc-900);
  --color-text-light: var(--zinc-700);
  --color-text-light-1: var(--zinc-650);
  --color-text-light-2: var(--zinc-600);
  --color-text-light-3: var(--zinc-550);
  --color-footer: var(--zinc-100);
  --color-timeline: var(--zinc-200);
  --color-input-text: var(--zinc-800);
  --color-input-background: #fff;
  --color-input-toggle-background: #fff;
  --color-input-border: var(--zinc-300);
  --color-input-border-hover: var(--zinc-400);
  --color-header-wrapper: var(--zinc-50);
  --color-header-wrapper-transparent: #d2e0f000;
  --color-light: #ffffffcc;
  --color-light-mimic-enabled: rgba(0, 0, 0, calc(6 / 255 * 222 / 255 / var(--opacity-disabled)));
  --color-light-border: #0000001d;
  --color-hover: #e4e4e4aa;
  --color-active: #e2e2e5;
  --color-menu: var(--zinc-100);
  --color-card: var(--zinc-50);
  --fancy-card-bg: var(--zinc-100);
  --fancy-card-border: var(--zinc-200);
  --color-markup-table-row: #ffffff06;
  --color-markup-code-block: var(--zinc-150);
  --color-markup-code-inline: var(--zinc-200);
  --color-button: var(--zinc-150);
  --color-code-bg: var(--zinc-50);
  --color-shadow: #00000060;
  --color-secondary-bg: var(--zinc-100);
  --color-text-focus: #fff;
  --color-expand-button: var(--zinc-200);
  --color-placeholder-text: var(--color-text-light-3);
  --color-editor-line-highlight: var(--zinc-100);
  --color-project-board-bg: var(--color-secondary-light-2);
  --color-project-board-dark-label: var(--color-text-light-3);
  --color-caret: var(--color-text);
  /* should ideally be --color-text-dark, see #15651 */
  --color-reaction-bg: #0000000a;
  --color-reaction-active-bg: var(--color-primary-alpha-20);
  --color-reaction-hover-bg: var(--color-primary-alpha-30);
  --color-tooltip-text: #ffffff;
  --color-tooltip-bg: #000000f0;
  --color-nav-bg: var(--zinc-100);
  --color-nav-hover-bg: var(--zinc-300);
  --color-secondary-nav-bg: var(--color-body);
  --color-label-text: var(--color-text);
  --color-label-bg: #cacaca7b;
  --color-label-hover-bg: #cacacaa0;
  --color-label-active-bg: #cacacaff;
  --color-label-bg-alt: #cacacaff;
  --color-accent: var(--color-primary-light-1);
  --color-small-accent: var(--color-primary-light-5);
  --color-highlight-fg: var(--color-primary-light-4);
  --color-highlight-bg: var(--color-primary-light-6);
  --color-overlay-backdrop: #080808c0;
  --color-selection-bg: var(--steel-450);
  --color-selection-fg: var(--color-white);
  /* pattern colors for gradient */
  --checkerboard-color-1: #ffffff;
  --checkerboard-color-2: #e5e5e5;
  --color-indicator-offline: #52525b;
  --color-indicator-offline-20: #52525b33;
  --color-indicator-idle: #16a34a;
  --color-indicator-idle-20: #16a34a33;
  --color-indicator-active: #2185d0;
  --color-indicator-active-20: #2185d033;
  accent-color: var(--color-accent);
  color-scheme: light;
}
.ui.secondary.vertical.menu {
  border-radius: 0.28571429rem !important;
  overflow: hidden;
}
.ui.basic.primary.button.item {
  background-color: var(--color-active) !important;
  color: var(--color-text) !important;
  box-shadow: none !important;
}
.ui.red.label.notification_count,
.ui.primary.labels .label {
  background-color: var(--color-primary-dark-1) !important;
}
.ui.labeled.icon.buttons > .button > .icon,
.ui.labeled.icon.button > .icon {
  background-color: var(--color-shadow) !important;
}
#review-box .review-comments-counter {
  background-color: var(--color-label-bg) !important;
  margin-left: 0.5em;
}
.ui.basic.labels .primary.label,
.ui.ui.ui.basic.primary.label {
  color: var(--color-text-dark) !important;
}
.ui.yellow.label.pending-label {
  background: var(--color-warning-bg) !important;
  color: var(--color-text-dark) !important;
}
