@import "./theme-forgejo-light.css";

:root {
  --color-diff-removed-word-bg: #c8c850;
  --color-diff-removed-row-border: #c8c850;
  --color-diff-removed-row-bg: #ffecc4;
  --color-diff-added-word-bg: #b8c0ff;
  --color-diff-added-row-border: #b8c0ff;
  --color-diff-added-row-bg: #e0e0ff;
  --color-code-bg: #ffffff;
}
