@import "tributejs/dist/tribute.css";

.tribute-container {
  box-shadow: 0 0.25rem 0.5rem rgba(0, 0, 0, 0.25);
  border-radius: var(--border-radius);
}

.tribute-container ul {
  margin-top: 0 !important;
  background: var(--color-body) !important;
}

.tribute-container li {
  padding: 3px 0.5rem !important;
}

.tribute-container li span.fullname {
  font-weight: var(--font-weight-normal);
  font-size: 0.8rem;
  margin-left: 3px;
}

.tribute-container li.highlight,
.tribute-container li:hover {
  background: var(--color-primary) !important;
  color: var(--color-primary-contrast) !important;
}

.tribute-item {
  display: flex;
  align-items: center;
}

.tribute-item .emoji,
.tribute-item img[src*="/avatar/"] {
  margin-right: 0.5rem;
}

.tribute-container img {
  width: 1.5rem !important;
  height: 1.5rem !important;
}
