.js-color-picker-input {
  display: flex;
  position: relative;
}

.js-color-picker-input input {
  padding-top: 8px !important;
  padding-bottom: 8px !important;
  padding-left: 32px !important;
}

.js-color-picker-input .preview-square {
  position: absolute;
  aspect-ratio: 1;
  height: 16px;
  left: 10px;
  top: 50%;
  transform: translateY(-50%);
  border-radius: 2px;
  background: repeating-linear-gradient(45deg, #aaa 25%, transparent 25%, transparent 75%, #aaa 75%, #aaa), repeating-linear-gradient(45deg, #aaa 25%, #fff 25%, #fff 75%, #aaa 75%, #aaa); /* stylelint-disable-line scale-unlimited/declaration-strict-value */
  background-position: 0 0, 4px 4px;
  background-size: 8px 8px;
}

.js-color-picker-input .preview-square::after {
  content: "";
  position: absolute;
  width: 100%;
  height: 100%;
  border-radius: inherit;
  background-color: currentcolor;
}

hex-color-picker {
  width: 180px;
  height: 120px;
}

hex-color-picker::part(hue-pointer),
hex-color-picker::part(saturation-pointer) {
  width: 22px;
  height: 22px;
}

hex-color-picker::part(hue) {
  flex-basis: 16px;
}
