.breadcrumb {
  display: flex;
  flex-wrap: wrap;
  align-items: center;
  gap: 3px;
}

.breadcrumb .breadcrumb-divider {
  color: var(--color-text-light-2);
}

.breadcrumb > * {
  display: inline;
}
