// Copyright 2021 The Gitea Authors. All rights reserved.
// SPDX-License-Identifier: MIT

package v1_16

// We used to use a table `remote_version` to store information for updater, now we use `AppState`, so this migration task is a no-op now.
