*,
*::before,
*::after {
  box-sizing: border-box;
}

body {
  margin: 0;
  background: #fff;
}

.swagger-back-link {
  color: #1f69c0;
  text-decoration: none;
  position: absolute;
  top: 1rem;
  right: 1.5rem;
  display: flex;
  align-items: center;
}

.swagger-back-link:hover {
  text-decoration: underline;
}

.swagger-back-link svg {
  color: inherit;
  fill: currentcolor;
  margin-right: 0.5rem;
}

@media (prefers-color-scheme: dark) {
  body {
    background: #1e1e1e;
  }
  .swagger-ui, .swagger-back-link {
    filter: invert(88%) hue-rotate(180deg);
  }
  .swagger-ui .microlight {
    filter: invert(100%) hue-rotate(180deg);
  }
}
