.file-view tr.active {
  color: inherit !important;
  background: inherit !important;
}

.lines-escape {
  width: 0;
}

.lines-code {
  padding-left: 5px;
}

.file-view tr.active .lines-num,
.file-view tr.active .lines-escape,
.file-view tr.active .lines-code {
  background: var(--color-highlight-bg) !important;
}

.file-view tr.active:last-of-type .lines-code {
  border-bottom-right-radius: var(--border-radius);
}

.file-view tr.active .lines-num {
  position: relative;
}

.file-view tr.active .interact-bg:hover {
  background: var(--color-primary-alpha-50) !important;
}

.file-view tr.active .lines-num::before {
  content: "";
  position: absolute;
  left: 0;
  width: 2px;
  height: 100%;
  background: var(--color-highlight-fg);
}
