.migrate .svg.gitea-git {
  --git-logo-color: #f05133;
  color: var(--git-logo-color);
}

.migrate-entries {
  display: grid;
  /* Limited to 4 cols by 1280px container */
  grid-template-columns: repeat(auto-fit, minmax(20rem, 1fr));
  gap: 1.5rem;
}

.migrate-entry svg {
  padding: 1.5rem;
}

.migrate-entry {
  display: flex;
  flex-direction: column;
  color: var(--color-text);
  background: var(--color-card);
  border: 1px solid var(--fancy-card-border);
  border-radius: var(--border-radius-large);
  transition: all 0.1s ease-in-out;
}

.migrate-entry:hover {
  transform: scale(105%);
  box-shadow: 0 0.5rem 1rem var(--color-shadow);
  color: var(--color-text);
}

.migrate-entry .content {
  width: 100%;
  margin-top: .5rem;
  padding: 1rem;
  flex: 1;
}

.migrate-entry .description {
  margin-top: .5rem;
  text-wrap: balance;
}

/* Desktop layout features */
@media (min-width: 599.98px) {
  .migrate-entry .content {
    text-align: center;
    border-top: 1px solid var(--fancy-card-border);
    border-radius: 0 0 var(--border-radius-large) var(--border-radius-large);
    background: var(--fancy-card-bg);
  }
}

/* Mobile layout features */
@media (max-width: 600px) {
  .migrate-entries {
    grid-template-columns: repeat(1, 1fr);
  }
  .migrate-entry {
    flex-direction: row;
  }
  .migrate-entry svg {
    height: 100%;
    width: 100%;
    max-width: 128px;
  }
}
