.repository.file.editor .tab[data-tab="write"] {
  padding: 0 !important;
}

.repository.file.editor .tab[data-tab="write"] .editor-toolbar {
  border: 0 !important;
}

.repository.file.editor .tab[data-tab="write"] .CodeMirror {
  border-left: 0;
  border-right: 0;
  border-bottom: 0;
}

.repo-editor-header {
  display: flex;
  margin: 1rem 0;
  padding: 3px 0;
}

.editor-toolbar {
  border-color: var(--color-secondary);
}

.editor-toolbar.fullscreen {
  background: var(--color-body);
}

.editor-toolbar button {
  border: none !important;
  color: var(--color-text-light);
}

.editor-toolbar button:not(:hover) {
  background-color: transparent !important;
}

.editor-toolbar i.separator {
  border-left: none;
  border-right-color: var(--color-secondary);
}

.editor-toolbar button:hover {
  background: var(--color-hover);
}

.editor-toolbar button.active {
  background: var(--color-active);
}

/* hide preview button, we have the preview tab for this */
.editor-toolbar:not(.fullscreen) .preview {
  display: none;
}

/* hide revert button in fullscreen, it breaks the page */
.editor-toolbar.fullscreen .revert-to-textarea {
  display: none;
}

.editor-preview {
  background-color: var(--color-body);
}

.editor-preview-side {
  border-color: var(--color-secondary);
}

.editor-statusbar {
  color: var(--color-text-light);
}

.editor-loading {
  padding: 1rem;
  text-align: center;
}

.edit-diff {
  padding: 0 !important;
}

.edit-diff > div > .ui.table {
  border-top: none !important;
  border-bottom: none !important;
}
